/* Replacement for broken version in GSL release. */

#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>

#define isfinite(x) ((x) <= (x) && (x) + 1e308 != (x))

double gsl_ldexp (double x, int e) {
  double factor = (e >= 0) ? 2.0 : 0.5;
  unsigned int remaining = (e >= 0) ? e : -e;
  while (remaining != 0) {
    if (remaining & 0x1) x *= factor;
    double newFactor = factor * factor;
    if (isfinite(newFactor)) factor = newFactor;
    else {
      if (remaining < 2) return x;
      else if (remaining == 2) return (x * factor) * factor;
      else if (remaining == 3) return ((x * factor) * factor) * factor;
      else if (remaining == 4)
	return (((x * factor) * factor) * factor) * factor;
      else return x * newFactor;
    }
    remaining >>= 1;
  }
  return x;
}

double gsl_frexp (double x, int *e) {
  if (!isfinite(x) || x == 0.0) { *e = 0; return x; }
  double ex = ceil(log(fabs(x))/M_LN2);
  int ei    = (int)ex;
  double f  = gsl_ldexp(x, -ei);

  if (!isfinite(f)) {
    *e = 0;
    GSL_ERROR_VAL
      ("in gsl_frexp: gsl_ldexp unexpectedly returned nonfinite value",
       GSL_ESANITY, x);
  }
  while (fabs(f) >= 1.0) { ++ei; f *= 0.5; }
  while (fabs(f) < 0.5) { --ei; f *= 2.0; }
  *e = ei;
  return f;
}
